# `atropy`: Wrapper for the `atropy_core` C++ library

## Objective
`atropy` is a python wrapper for `atropy_core`, which solves the chemical master equation (CME) and is described further at https://gitlab.com/atropy/atropy_core.git.

`atropy` allows the user to build models and run simulations, using the functions in `atropy/core/generator.py`. Examples of use can be found in `atropy/src/examples` or also on the project website https://atropy.gitlab.io/.

## Requirements
- CMake (3.27 or later)
- C++20 compatible C++ compiler
- Fortran compiler (if OpenBLAS is used)
- Eigen 3.4
- HDF5 (1.10.x)
- Intel MKL
- netCDF4
- OpenMP
- Python (>=3.10)

## Installation

The package can be found on PyPI and the installation can thus be exuecuted using uv by astral or the classical pip download. We recommend the approach using uv, due to faster installation times.

### uv installation

Generate a uv virtual environment by executing
```shell
uv venv <environment_name> --python <python_version>
```

The environment can be found in the current shell directory.

Next activate the environment with
```shell
source <path_to_environment>/bin/activate
```

Now one can download `atropy` using the python packaging index
```shell
uv pip install atropy
```

The package is installed in `<path_to_environment>/lib/python<python_version>/site-packages` with all the dependencies, which were also downloaded from PyPI.

### pip installation

Using classical pip installation execute
```shell
pip install atropy
```

Again using a virtual environment is still recommended.

## Example problems

After installation several example problems can be run. Those can be found in `<path_to_atropy>/src/examples`. One can run a problem by executing
```shell
python3 <path_to_atropy>/src/examples/run_lambda_phage.py
```

If successful, output files should be generated in the current working directory.